import os
from anyio import Path

from material_zui.list import map_to, filter_to
from .type import ZuiFile


def get_names(directory_path: str) -> list[str]:
    '''
    Get all file names in a directory
    '''
    return os.listdir(directory_path)


def get_paths(directory_path: str) -> list[str]:
    filenames = get_names(directory_path)
    return [os.path.join(directory_path, filename) for filename in filenames]


def get_file_info(file_path: str) -> ZuiFile:
    """
    The function takes a file path as input and returns information about the file such as its name,
    extension, and directory name.

    @param file_path: A string representing the path to a file
    @type file_path: str
    @return: A dictionary containing information about the file specified by the input file path. The
    dictionary includes the file name, name (without extension), directory name, and file extension.
    """
    path = Path(file_path)
    file_name = path.name
    name = path.stem
    ext = os.path.splitext(file_path)[1][1:]
    dir_name = os.path.dirname(file_path)
    return {'file_name': file_name, 'name': name, 'dir_name': dir_name, 'ext': ext}


# def get_files_info(directory_path: str) -> list[ZuiFile]:
#     return map_to(get_paths(directory_path), lambda file_path, _: get_file_info(file_path))
def get_files_info(directory_path: str, ext: str = '') -> list[ZuiFile]:
    """
    Returns a list of file information for all files in the specified directory path with the given file extension (if provided). 

    :param directory_path: A string representing the path to the directory to search for files.
    :param ext: An optional string representing the file extension to filter files by.
    :return: A list of ZuiFile objects containing information about each file in the directory (if no extension provided) or only files with the specified extension.
    """
    files_info = map_to(get_paths(directory_path),
                        lambda file_path, _: get_file_info(file_path))
    return files_info if ext == '' else filter_to(files_info, lambda file_info, _: file_info['ext'] == ext)



from .common import get_names, get_file_info, get_files_info
from .download import download
from .excel import write_excel
from .read import read_file_to_list, load_json_array, load_json_object, load_diff_line
from .type import ZuiFile, ZuiExcelColumn, ZuiExcelColumns, ZuiExcelDataItem, ZuiExcelData
from .write import write_json, write_to_last, write



import cv2
# import matplotlib.pyplot as plt
# Acquire a sample image and specify its current size:
# read image
img = cv2.imread("my-image.jpg")
print('Image Width is', img.shape[1])
print('Image Height is', img.shape[0])
# Resize the image of, say, a size of 800×600 pixels, to 300×300 pixels:
imgResize = cv2.resize(img, (300, 300))
print('end', img.shape)

img = cv2.imread('my-image.jpg', cv2.IMREAD_GRAYSCALE)
# cv2.imshow('image', img)
cv2.imshow('image', imgResize)
cv2.imwrite('img/img2.jpg', imgResize)

cv2.waitKey(0)
cv2.destroyAllWindows()



from typing import Any


def to_float(input_value: Any) -> bool:
    try:
        return float(input_value)
    except ValueError:
        return False


def is_int(value: Any) -> bool:
    '''
    use `isinstance(value, int)` not work for case value is `False`
    '''
    return type(value) is int


def is_float(value: Any) -> bool: return type(value) is float



from collections import defaultdict
import json
from typing import DefaultDict
from pyparsing import Any

from .write import write_to_last
from material_zui.list import get_diff
from material_zui.string import remove_all
from material_zui.utility import pipe_list


def read_file_to_list(filename: str) -> list[str]:
    with open(filename, "r") as f:
        lines = f.readlines()
    return pipe_list(remove_all('\n'))(lines)


def load_json_object(json_file_path: str) -> DefaultDict[str, str]:
    '''
    Load json file data to dict type
    @json_file_path: json file path
    @return: dict json data
    '''
    with open(json_file_path, encoding='utf-8') as json_data:
        return defaultdict(str, json.load(json_data))


def load_json_array(json_file_path: str) -> list[dict[Any, Any]]:
    '''
    Load json file data to list type
    @json_file_path: json file path
    @return: list json data
    '''
    with open(json_file_path, "r") as f:
        data = json.load(f)
    return list(data)


def read_json(json_file_path: str):
    '''
    Read json file and return content value
    @return: `dict` for json object, otherwise `list dict` for json array
    - ex: using: `data[0]` or `data['field_value']`
    '''
    return json.loads(open(json_file_path, encoding="utf-8").read())


def load_diff_line(file_urls_path: str, input_urls: list[str], is_write_file: bool = False) -> list[str]:
    saved_urls = read_file_to_list(file_urls_path)
    diff_urls = get_diff(input_urls, saved_urls)
    print("New lines", len(diff_urls))
    print("Duplicate lines", len(input_urls)-len(diff_urls))
    if is_write_file:
        write_to_last(file_urls_path, '\n---New lines---\n' +
                      "\n".join(diff_urls))
    return diff_urls



import re

# The string we want to search
# text = '''{"role":"assistant","id":"chatcmpl-7BcGlt1VxezpfB3C8lJDZUmTq0nNp","parentMessageId":"00b954a2-da94-4c07-9641-4ff6b702d6f5","text":"","detail":{"id":"chatcmpl-7BcGlt1VxezpfB3C8lJDZUmTq0nNp","object":"chat.completion.chunk","created":1683003071,"model":"gpt-3.5-turbo-0301","choices":[{"delta":{"role":"assistant"},"index":0,"finish_reason":null}]}}
# {"role":"assistant","id":"chatcmpl-7BcGlt1VxezpfB3C8lJDZUmTq0nNp","parentMessageId":"00b954a2-da94-4c07-9641-4ff6b702d6f5","text":"Hello","delta":"Hello","detail":{"id":"chatcmpl-7BcGlt1VxezpfB3C8lJDZUmTq0nNp","object":"chat.completion.chunk","created":1683003071,"model":"gpt-3.5-turbo-0301","choices":[{"delta":{"content":"Hello"},"index":0,"finish_reason":null}]}}
# {"role":"assistant","id":"chatcmpl-7BcGlt1VxezpfB3C8lJDZUmTq0nNp","parentMessageId":"00b954a2-da94-4c07-9641-4ff6b702d6f5","text":"Hello again","delta":" again","detail":{"id":"chatcmpl-7BcGlt1VxezpfB3C8lJDZUmTq0nNp","object":"chat.completion.chunk","created":1683003071,"model":"gpt-3.5-turbo-0301","choices":[{"delta":{"content":" again"},"index":0,"finish_reason":null}]}}
# {"role":"assistant","id":"chatcmpl-7BcGlt1VxezpfB3C8lJDZUmTq0nNp","parentMessageId":"00b954a2-da94-4c07-9641-4ff6b702d6f5","text":"Hello again!","delta":"!","detail":{"id":"chatcmpl-7BcGlt1VxezpfB3C8lJDZUmTq0nNp","object":"chat.completion.chunk","created":1683003071,"model":"gpt-3.5-turbo-0301","choices":[{"delta":{"content":"!"},"index":0,"finish_reason":null}]}}
# {"role":"assistant","id":"chatcmpl-7BcGlt1VxezpfB3C8lJDZUmTq0nNp","parentMessageId":"00b954a2-da94-4c07-9641-4ff6b702d6f5","text":"Hello again! How","delta":" How","detail":{"id":"chatcmpl-7BcGlt1VxezpfB3C8lJDZUmTq0nNp","object":"chat.completion.chunk","created":1683003071,"model":"gpt-3.5-turbo-0301","choices":[{"delta":{"content":" How"},"index":0,"finish_reason":null}]}}
# {"role":"assistant","id":"chatcmpl-7BcGlt1VxezpfB3C8lJDZUmTq0nNp","parentMessageId":"00b954a2-da94-4c07-9641-4ff6b702d6f5","text":"Hello again! How can","delta":" can","detail":{"id":"chatcmpl-7BcGlt1VxezpfB3C8lJDZUmTq0nNp","object":"chat.completion.chunk","created":1683003071,"model":"gpt-3.5-turbo-0301","choices":[{"delta":{"content":" can"},"index":0,"finish_reason":null}]}}
# {"role":"assistant","id":"chatcmpl-7BcGlt1VxezpfB3C8lJDZUmTq0nNp","parentMessageId":"00b954a2-da94-4c07-9641-4ff6b702d6f5","text":"Hello again! How can I","delta":" I","detail":{"id":"chatcmpl-7BcGlt1VxezpfB3C8lJDZUmTq0nNp","object":"chat.completion.chunk","created":1683003071,"model":"gpt-3.5-turbo-0301","choices":[{"delta":{"content":" I"},"index":0,"finish_reason":null}]}}
# {"role":"assistant","id":"chatcmpl-7BcGlt1VxezpfB3C8lJDZUmTq0nNp","parentMessageId":"00b954a2-da94-4c07-9641-4ff6b702d6f5","text":"Hello again! How can I help","delta":" help","detail":{"id":"chatcmpl-7BcGlt1VxezpfB3C8lJDZUmTq0nNp","object":"chat.completion.chunk","created":1683003071,"model":"gpt-3.5-turbo-0301","choices":[{"delta":{"content":" help"},"index":0,"finish_reason":null}]}}
# {"role":"assistant","id":"chatcmpl-7BcGlt1VxezpfB3C8lJDZUmTq0nNp","parentMessageId":"00b954a2-da94-4c07-9641-4ff6b702d6f5","text":"Hello again! How can I help you","delta":" you","detail":{"id":"chatcmpl-7BcGlt1VxezpfB3C8lJDZUmTq0nNp","object":"chat.completion.chunk","created":1683003071,"model":"gpt-3.5-turbo-0301","choices":[{"delta":{"content":" you"},"index":0,"finish_reason":null}]}}
# {"role":"assistant","id":"chatcmpl-7BcGlt1VxezpfB3C8lJDZUmTq0nNp","parentMessageId":"00b954a2-da94-4c07-9641-4ff6b702d6f5","text":"Hello again! How can I help you today","delta":" today","detail":{"id":"chatcmpl-7BcGlt1VxezpfB3C8lJDZUmTq0nNp","object":"chat.completion.chunk","created":1683003071,"model":"gpt-3.5-turbo-0301","choices":[{"delta":{"content":" today"},"index":0,"finish_reason":null}]}}
# {"role":"assistant","id":"chatcmpl-7BcGlt1VxezpfB3C8lJDZUmTq0nNp","parentMessageId":"00b954a2-da94-4c07-9641-4ff6b702d6f5","text":"Hello again! How can I help you today?","delta":"?","detail":{"id":"chatcmpl-7BcGlt1VxezpfB3C8lJDZUmTq0nNp","object":"chat.completion.chunk","created":1683003071,"model":"gpt-3.5-turbo-0301","choices":[{"delta":{"content":"?"},"index":0,"finish_reason":null}]}}
# {"role":"assistant","id":"chatcmpl-7BcGlt1VxezpfB3C8lJDZUmTq0nNp","parentMessageId":"00b954a2-da94-4c07-9641-4ff6b702d6f5","text":"Hello again! How can I help you today?","detail":{"id":"chatcmpl-7BcGlt1VxezpfB3C8lJDZUmTq0nNp","object":"chat.completion.chunk","created":1683003071,"model":"gpt-3.5-turbo-0301","choices":[{"delta":{},"index":0,"finish_reason":"stop"}]}}
# '''

# # The regular expression we want to use
# regex = '.+"finish_reason":"stop".+'

# # Create a regular expression object
# pattern = re.compile(regex)

# # Search the string for the regular expression
# match = pattern.search(text)

# # Check if the match was found
# if match:
#     # The match was found, so print the matched text
#     json_data = json.loads(match.group())
#     print(json_data['text'])
# else:
#     # The match was not found, so print a message
#     print("No match found")


def check(value: str, regex: str):
    pattern = re.compile(regex)
    return pattern.search(value)


def value(value: str, regex: str) -> str:
    match = check(value, regex)
    return match.group() if match else ''



from pathlib import Path
from time import sleep
from typing import Any
from EdgeGPT.EdgeGPT import Chatbot, ConversationStyle
from EdgeGPT.EdgeUtils import Query, ImageQuery, Cookie

from material_zui.file import write_excel, ZuiExcelColumns
from .result import ZuiBingAiResult
from .type import ZuiArticle


class ZuiBingAi:
    creative = ConversationStyle.creative
    balanced = ConversationStyle.balanced
    precise = ConversationStyle.precise

    def __init__(self, directory_cookie_path: str = '') -> None:
        '''
        - Base on https://pypi.org/project/EdgeGPT
        @directory_cookie_path: `directory path` of cookie file (file cookie must have pattern `bing_cookies_*.json`)
        - In case only ask, DO NOT NEED to login for regions allow using free Bing chat AI (support `Vietnam`), but sometimes reseponse required login
        - If meet error `EdgeGPT.exceptions.NotAllowedToAccess: Sorry, you need to login first to access this service.`, just try again
        '''
        self.response: dict[Any, Any] = {}
        if directory_cookie_path:
            self.set_cookie(directory_cookie_path)

    def set_cookie(self, directory_cookie_path: str) -> None:
        self.directory_cookie_path = Path(directory_cookie_path).resolve()
        Cookie.dirpath = self.directory_cookie_path

    async def create(self):
        self.bot = await Chatbot.create()

    async def ask(self, prompt: str, conversation_style: ConversationStyle = ConversationStyle.creative) -> ZuiBingAiResult:
        '''
        In case only ask, DO NOT NEED to login for regions allow using free Bing chat AI (support `Vietnam`)
        '''
        if not hasattr(self, 'bot'):
            await self.create()
        self.response = await self.bot.ask(prompt=prompt, conversation_style=conversation_style)
        return ZuiBingAiResult(self.response)

    def query(self, prompt: str, style: str = "balanced", directory_cookie_path: str = ''):
        '''
        @style: including `creative`, `balanced`, `precise`
        @directory_cookie_path: directory of cookie file (file cookie must have pattern `bing_cookies_*.json`)
        '''
        if directory_cookie_path:
            self.set_cookie(directory_cookie_path)
        self.responseQuery = Query(prompt, style=style)
        self.response = self.responseQuery.response  # type: ignore
        return ZuiBingAiResult(self.response)  # type: ignore

    def gen_image(self, prompt: str):
        '''
        Must input cookie to use this feature
        '''
        # Cookie.image_dirpath
        return ImageQuery(prompt)

    def gen_article(self, topics: list[str]) -> list[ZuiArticle]:
        data = []
        for index, topic in enumerate(topics):
            find_keywords = f'Identify 10 SEO keywords related to {topic}'
            keywords = self.query(find_keywords)
            data.append({
                'title': index,
                'description': 'description',
                'content': 'content',
                'keywords': keywords.last_text,
            })
            sleep(200)
        return data

    def article_to_excel(self, file_path: str, topics: list[str]) -> None:
        columns: ZuiExcelColumns = [
            {'field': 'title', 'name': 'title'},
            {'field': 'description', 'name': 'Description'},
            {'field': 'content', 'name': 'Content'},
            {'field': 'keywords', 'name': 'Keywords'},
        ]
        data: list[Any] = self.gen_article(topics)
        write_excel(file_path, columns, data)



import cv2
from cv2 import Mat
from material_zui.image.common import get_image_size
from material_zui.image.data import MAX_MP

# def upscale(path:  str | bytes, newPath: str, wRatio: int = 1, hRatio: int = 1, newWidth: int = 0, newHeight: int = 0, isShowImage: bool = False):
#     """
#     Use AI to upscale image
#     """
#     image = Image.open(path)
#     toWidth, toHeight = [0, 0]
#     if newWidth != 0 and newHeight != 0:
#         toWidth = newWidth
#         toHeight = newHeight
#     else:
#         width, height = image.size
#         toWidth = width*wRatio
#         toHeight = height*hRatio
#     newImage = image.resize((toWidth, toHeight))
#     newImage.save(newPath)
#     print(image.format)
#     if isShowImage:
#         newImage.show()
#     return


# def upscale_to_size(image: ZuiImage, new_width: int = 0,
#                     new_height: int = 0): return image.resize((new_width, new_height))


# def upscale_to_ratio(image: ZuiImage, w_ratio: int = 1, h_ratio: int = 1):
#     width, height = image.size
#     to_width, to_height = [width*w_ratio, height*h_ratio]
#     return image.resize((to_width, to_height))


# def upscales_to_max_size(image: ZuiImage, max_mp: int = MAX_MP) -> ZuiImage:
#     '''
#     @max_mp: max size in megapixels
#     '''
#     width, height = image.size
#     to_width, to_height = image.size
#     while (to_width*to_height < max_mp):
#         to_width += width
#         to_height += height
#     return image.resize((to_width, to_height))


# def upscale_to_max_size(images: list[dict[{'image': ZuiImage, 'max_mp': int}]]): return list(map(lambda image: upscales_to_max_size(
#     image['image'], image['max_mp']), images))


# def save_to(images: list[ZuiImage], path: str):
#     for index, image in enumerate(images):
#         # new_path = os.path.join(path, str(index)+'.jpg')
#         # new_path = os.path.join(path, "{}.{}".format(index, image.format))
#         new_path = os.path.join(path, "{}.{}".format(index, 'jpg'))
#         print(new_path, image.getim())
#         image.save(new_path)

def get_ratio_to_max_mp(width: int, height: int, max_mp: int = MAX_MP) -> int:  # type: ignore
    '''
    @max_mp: max size in megapixels
    '''
    to_width, to_height = width, height
    ratio = 1
    while (to_width*to_height < max_mp):
        ratio += 1
        to_width = width * ratio
        to_height = height * ratio
    return ratio


# @dispatch(Mat, int)
# def get_ratio_to_max_mp(image: Mat, max_mp: int = MAX_MP) -> int:
#     '''
#     @max_mp: max size in megapixels
#     '''
#     print('image', image)
#     height, width = image.shape[:2]
#     return get_ratio_to_max_mp(width, height, max_mp)  # type: ignore

# def get_image_ratio_to_max_mp(image: Mat, max_mp: int = MAX_MP) -> int:
#     '''
#     @max_mp: max size in megapixels
#     '''
#     height, width = image.shape[:2]
#     return get_ratio_to_max_mp(width, height, max_mp)  # type: ignore


def upscale_to_max_size(max_mp: int = MAX_MP):
    def handle_image(image: Mat):
        height, width = image.shape[:2]
        width, height = get_image_size(image)
        ratio = get_ratio_to_max_mp(width, height, max_mp)
        upscale_image = cv2.resize(
            image, None, fx=ratio, fy=ratio, interpolation=cv2.INTER_CUBIC)  # type: ignore
        return upscale_image
    return handle_image


# def upscales_dir_to_max_size(directory_path:  str, new_directory_path: str = '', max_mp: int = MAX_MP) -> list[ZuiImage]:
#     '''
#     @max_mp: max size in megapixels
#     '''
#     new_directory_path = new_directory_path if new_directory_path else directory_path
#     images_name = get_image_names(directory_path)
#     images: list[ZuiImageRes] = []
#     for image_name in images_name:
#         image = cv2.imread(os.path.join(directory_path, image_name))
#         height, width = image.shape[:2]
#         ratio = get_ratio_to_max_mp(width, height, max_mp)  # type: ignore
#         upscale_image = cv2.resize(
#             image, None, fx=ratio, fy=ratio, interpolation=cv2.INTER_CUBIC)  # type: ignore
#         images.append({'name': image_name, 'width': width,
#                       'height': height, 'matImage': upscale_image, 'ratio': ratio})
#     return images  # type: ignore


# def save_upscales_dir_to_max_size(directory_path:  str, new_directory_path: str = '', max_mp: int = MAX_MP):
#     '''
#     @max_mp: max size in megapixels
#     '''
#     images = upscales_dir_to_max_size(
#         directory_path, new_directory_path, max_mp)
#     printTable({
#         'Name': [info['name'] for info in images],
#         'Width': [info['width'] for info in images],
#         'Height': [info['height'] for info in images],
#         'Ratio': [info['ratio'] for info in images],
#     })



from typing import TypeVar, Callable, Any

T = TypeVar('T')

R = TypeVar('R')


def is_last_index(list: list[Any], index: int) -> bool:
    return index == len(list) - 1


def list_range(list: list[T], limit: int = 0, start_index: int = 0) -> list[T]:
    end_index = start_index+limit
    return list[start_index:end_index] if end_index else list


# def map_to(items: list[T], handle_function: Callable[[T, int], R]) -> list[R]:
#     return list(map(handle_function, items, range(len(items))))

def map_to(items: list[T], handle_function: Callable[[T, int], R]) -> list[R]:
    return list(map(handle_function, items, range(len(items))))

# def map_to(items: List[Match[str]], transformer: Callable[[Match[str], int], str]) -> List[str]:
#     return [transformer(item, i) for i, item in enumerate(items)]


def filter_to(items: list[T], handle_function: Callable[[T, int], bool]) -> list[T]:
    result: list[T] = []
    for index, item in enumerate(items):
        if handle_function(item, index):
            result.append(item)
    return result

# def filter(items: list[T], handle_function: Callable[[T], bool]) -> list[T]:
#     return list(filter(handle_function, items))


def get_diff(list1: list[T], list2: list[T]) -> list[T]:
    """Get the difference between two lists.
    Args:
        list1 (list): The first list.
        list2 (list): The second list.
    Returns:
        list: item in list1 not in list2
    """
    return filter_to(list1, lambda item, _: item not in list2)


def get(index: int, default_value: Any = None) -> Callable[[list[Any]], Any | None]:
    """
    Safely gets an element from a list.
    Args:
        list1: The list to be accessed.
        index: The index of the element to be retrieved.
    Returns:
        The element at index `index` in list1, or `None` if the index is out of bounds.
    """
    return lambda items: items[index] if index > -1 and index < len(items) else default_value

# def get(list1:list[Any], index:int):
#   """
#   Safely gets an element from a list.

#   Args:
#     list1: The list to be accessed.
#     index: The index of the element to be retrieved.

#   Returns:
#     The element at index `index` in list1, or `None` if the index is out of bounds.
#   """
#   if index < 0 or index >= len(list1):
#     return None
#   else:
#     return list1[index]

# numbers: list[int] = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
# def is_even(number: int) -> bool:
#     return number % 2 == 0
# even_numbers = filter(is_even, numbers)

# class ZuiList:
#     def __init__(self, items: list[T]) -> None:
#         self.list: list[T] = items

#     def map(self, handle_function: Callable[[T], R]) -> list[R]:
#         return list(map(handle_function, self.list))

#     def filter(self, items: list[T], handle_function: Callable[[T], bool]) -> list[T]:
#         return list(filter(handle_function, items))


# ZuiList([1, 2, 3]).map(lambda x: x+1)



from cv2 import Mat
import numpy as np
import cv2
import os

"""
Credits: 
	1. https://github.com/opencv/opencv/blob/master/samples/dnn/colorization.py
	2. http://richzhang.github.io/colorization/
	3. https://github.com/richzhang/colorization/

Download the model files:
	1. colorization_deploy_v2.prototxt: https://github.com/richzhang/colorization/tree/caffe/colorization/models
	2. pts_in_hull.npy: https://github.com/richzhang/colorization/blob/caffe/colorization/resources/pts_in_hull.npy
	3. colorization_release_v2.caffemodel: https://www.dropbox.com/s/dx0qvhhp5hbcx7z/colorization_release_v2.caffemodel?dl=1
"""

# Paths to load the models
DIR = f"{os.path.dirname(os.path.realpath(__file__))}/model"
PROTOTXT = os.path.join(DIR, "colorization_deploy_v2.prototxt")
POINTS = os.path.join(DIR, "pts_in_hull.npy")
MODEL = os.path.join(DIR, "colorization_release_v2.caffemodel")


def colorization(image: Mat) -> Mat:
    # Load the Model
    net = cv2.dnn.readNetFromCaffe(PROTOTXT, MODEL)
    pts = np.load(POINTS)

    # Load centers for ab channel quantization used for rebalancing.
    class8 = net.getLayerId("class8_ab")
    conv8 = net.getLayerId("conv8_313_rh")
    pts = pts.transpose().reshape(2, 313, 1, 1)
    net.getLayer(class8).blobs = [pts.astype("float32")]
    net.getLayer(conv8).blobs = [np.full([1, 313], 2.606, dtype="float32")]

    # Load the input image
    scaled = image.astype("float32") / 255.0
    lab = cv2.cvtColor(scaled, cv2.COLOR_BGR2LAB)
    resized = cv2.resize(lab, (224, 224))
    L = cv2.split(resized)[0]
    L -= 50

    # Colorizing the image
    net.setInput(cv2.dnn.blobFromImage(L))
    ab = net.forward()[0, :, :, :].transpose((1, 2, 0))
    ab = cv2.resize(ab, (image.shape[1], image.shape[0]))
    L = cv2.split(lab)[0]

    colorized = np.concatenate((L[:, :, np.newaxis], ab), axis=2)
    colorized = cv2.cvtColor(colorized, cv2.COLOR_LAB2BGR)
    colorized = np.clip(colorized, 0, 1)
    colorized = (255 * colorized).astype("uint8")  # type: ignore
    return colorized


def save_colorization(image_path: str, output_path: str):
    image = cv2.imread(image_path)
    colorized = colorization(image)
    cv2.imwrite(output_path, colorized)



import os


def file_exists(file_path: str) -> bool:
    """Checks if the file exists.
    Args:
      filename: The name of the file to check.
    Returns:
      True if the file exists, False otherwise.
    """
    return os.path.exists(file_path)


def directory_exists(directory_path: str) -> bool:
    """Checks if the directory exists.
    Args:
      path: The path of the directory to check.

    Returns:
      True if the directory exists, False otherwise.
    """
    return os.path.isdir(directory_path) and os.path.exists(directory_path)



from typing import TypeAlias, Any

ZuiFile = dict[{'name': str, 'ext': str, 'file_name': str, 'dir_name': str}]

ZuiExcelColumn: TypeAlias = dict[
    {
        'field': str,
        'name': str,
        # 'type': Optional[str],
        # 'width': Optional[str],
        # 'format': Optional[str]
    }
]

ZuiExcelColumns: TypeAlias = list[ZuiExcelColumn]

ZuiExcelDataItem = dict[str, Any]

ZuiExcelData = list[ZuiExcelDataItem]



from pyparsing import Any
from json import dump

from .check import file_exists


def write_json(data: dict[Any, Any] | list[Any], json_path: str, indent: int = 2) -> None:
    '''
    Write data to json file
    @data: is json object or array object
    @indent: space tab
    '''
    with open(json_path, "w") as f:
        dump(data, f, indent=indent)


def write_to_last(file_path: str, text: str) -> None:
    """
    Writes the given text to the last line of the file.
    @file_path: The path to the file.
    @text: The text to write.
    """
    if file_exists(file_path):
        with open(file_path, "r+") as f:
            f.seek(0, 2)
            f.write(text)
    else:
        write(file_path, text)


def write(file_path: str, text: str) -> None:
    """
    Writes an override file with the given content.
    @filename: The name of the override file.
    @content: The content to write to the override file.
    """
    with open(file_path, "w") as f:
        f.write(text)



import os
import platform
from typing import Any
from bs4 import BeautifulSoup
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.remote.webelement import WebElement

from material_zui.selenium.common import safe_find_element
from material_zui.fake import random_sleep
from material_zui.list import map_to


class Zui_Selenium_Chrome:
    def __init__(self) -> None:
        self.is_mac = False
        if not any(os_name in platform.platform() for os_name in ["Windows", "Linux"]):
            self.is_mac = True

    @property
    def screen_height(self) -> int:
        return self.driver.execute_script("return window.screen.height;")

    @property
    def scroll_height(self) -> int:
        return self.driver.execute_script("return document.body.scrollHeight;")

    @property
    def page_source(self) -> str: return self.driver.page_source

    @property
    def document(self) -> BeautifulSoup:
        return BeautifulSoup(self.driver.page_source, "html.parser")

    def connect(self):
        self.driver = webdriver.Chrome()
        return self.driver

    def connect_debug(self, port: int = 9000, debug_address: str = ''):
        '''
        @port: port number, default `9000`
        @debug_address: format `127.0.0.1:9000`, default `localhost:{port}`
        - Use for case need authorization, you just need to start `chrome beta` -> login account -> close browswer then
            - start `chrome` on debug mode -> call this method to connect to browser opened
        1. Install `chrome beta` for better automation: https://www.google.com/chrome/beta
        2. Start `chrome` by command: `google-chrome-beta --remote-debugging-port={port}`
                - `port` must the same with input parameter of this method
        '''
        debug_address = debug_address or f"localhost:{port}"
        self.options = Options()
        self.options.add_experimental_option(
            "debuggerAddress", debug_address)
        self.driver = webdriver.Chrome(options=self.options)
        return self.driver

    def delay(self, sec: float = 0) -> None:
        random_sleep(1, 5, sec)

    def safe_find_element(
            self, by: str = By.ID,
            value: str | None = None,
            parent: Any = None) -> WebElement | None:
        '''
        This is safe find element method
        @return `None` in case not found
        '''
        return safe_find_element(parent if parent else self.driver, by, value)

    def safe_find_element_by_xpath(self, xpath_value: str) -> WebElement | None:
        return self.safe_find_element(By.XPATH, xpath_value)

    def find_element_by_xpath(self, xpath_value: str): return self.driver.find_element(
        By.XPATH, xpath_value)

    def find_elements_by_xpath(self, xpath_value: str): return self.driver.find_elements(
        By.XPATH, xpath_value)

    def find_elements_by_class(self, class_value: str): return self.driver.find_elements(
        By.CLASS_NAME, class_value)

    def scroll_to_end(self) -> None:
        self.driver.execute_script("window.scrollTo(0, {scroll_height});".format(
            scroll_height=self.scroll_height))

    def scroll(self, step_scroll: int) -> None:
        for index in range(step_scroll):
            print(index+1, "Scrolling to", self.scroll_height)
            self.scroll_to_end()
            self.delay()

    def get_urls(self, class_selector: str) -> list[str]:
        '''
        Get video urls by class selector of each video
        '''
        videos = self.document.find_all(
            "div", {"class": class_selector})
        return map_to(videos, lambda video, _: video.a["href"])

    def switch_to_frame(self, xpath_value: str) -> None:
        '''
        @xpath_value: must be end with `frame/iframe` like `//*[@id="main"]/div[2]/div/iframe`
        '''
        frame_element = self.find_element_by_xpath(xpath_value)
        self.driver.switch_to.frame(frame_element)

    def upload_file(self, xpath_value: str, video_relative_path: str) -> None:
        '''
        @xpath_value: must be end with `input` like `//*[@id="root"]/div/div/div/div/div/div/div/input`
        @video_relative_path: video relative path, from root project directory
        '''
        file_input = self.find_element_by_xpath(xpath_value)
        abs_path = os.path.abspath(video_relative_path)
        file_input.send_keys(abs_path)

    def click(self, xpath_value: str):
        '''
        Click then return that element
        '''
        element = self.find_element_by_xpath(xpath_value)
        element.click()
        return element

    def safe_click(self, xpath_value: str):
        '''
        Safe click element
        '''
        try:
            element = self.safe_find_element_by_xpath(xpath_value)
            if element:
                element.click()
            return element
        except:
            return None

    def send_keys(self, xpath_value: str, key_value: str, clear_before_send_keys: bool = True, delay_time: int = 10):
        '''
        Send keys then return that element
        @xpath_value: only valid with element selector can input value
        @clear_before_send_keys:
        - True: clear input before send keys
        - False: value will input after existed value
        '''
        # element = self.find_element_by_xpath(xpath_value)
        element = self.wait_get(xpath_value, delay_time)
        element.click()
        if clear_before_send_keys:
            element.send_keys(Keys.CONTROL + 'a')
            element.send_keys(Keys.BACKSPACE)
        element.send_keys(key_value)
        return element

    # def wait_get(self, xpath_value: str, delay_time: int = 10) -> WebElement:
    #     element: Any = WebDriverWait(self.driver, delay_time).until(  # using explicit wait for 10 seconds
    #         EC.presence_of_element_located(
    #             (By.XPATH, xpath_value))  # finding the element
    #     )
    #     return element
    def wait_get(self, xpath_value: str, delay_time: int = 10, time_to_try: int = 10) -> WebElement:
        for _ in range(time_to_try):
            element = self.safe_find_element_by_xpath(xpath_value)
            if element:
                return element
            self.delay(delay_time)
        return WebDriverWait(self.driver, delay_time).until(  # using explicit wait for 10 seconds
            EC.presence_of_element_located(
                (By.XPATH, xpath_value))  # finding the element
        )



import replicate
import os


def set_token(token: str) -> None:
    '''
      Create account then get token from: https://replicate.com/account
    '''
    os.environ["REPLICATE_API_TOKEN"] = token


def run_stable_diffusion(prompt: str) -> list[str]:
    '''https://replicate.com/stability-ai/stable-diffusion'''
    return replicate.run(
        "stability-ai/stable-diffusion:27b93a2413e7f36cd83da926f3656280b2931564ff050bf9575f1fdf9bcd7478",
        input={"prompt": prompt},
    )  # type: ignore


def text_to_image(prompt: str) -> list[str]:
    '''https://replicate.com/pixray/text2image'''
    urls = replicate.run(
        "pixray/text2image:5c347a4bfa1d4523a58ae614c2194e15f2ae682b57e3797a5bb468920aa70ebf",
        input={"prompts": prompt},
    )
    return [url for url in urls]


def upscale(image_path: str, upscale: int = 2) -> str:
    '''https://replicate.com/sczhou/codeformer'''
    return replicate.run(
        "sczhou/codeformer:7de2ea26c616d5bf2245ad0d5e24f0ff9a6204578a5c876db53142edd9d2cd56",
        input={"image": open(image_path, "rb"), "upscale": upscale}
    )  # type: ignore


def recover_old_photo(image_path: str, with_scratch: bool = False, HR: bool = False) -> str:
    '''https://replicate.com/microsoft/bringing-old-photos-back-to-life'''
    return replicate.run(
        "microsoft/bringing-old-photos-back-to-life:c75db81db6cbd809d93cc3b7e7a088a351a3349c9fa02b6d393e35e0d51ba799",
        input={"image": open(image_path, "rb"),
               "with_scratch": with_scratch, "HR": HR}
    )  # type: ignore


def fill_color(image_path: str, classes: str = "88") -> list[dict[{'image': str}]]:
    '''https://replicate.com/cjwbw/bigcolor'''
    return replicate.run(
        "cjwbw/bigcolor:9451bfbf652b21a9bccc741e5c7046540faa5586cfa3aa45abc7dbb46151a4f7",
        input={"image": open(image_path, "rb"), "classes": classes}
    )  # type: ignore


def face_restore(image_path: str, task: str = "Face Restoration", broken_image: bool = False) -> list[str]:
    '''
    https://replicate.com/yangxy/gpen
    @task
    Allowed values: `Face Restoration`, `Face Colorization`, `Face Inpainting`
    Default value: `Face Restoration`
    '''
    return replicate.run(
        "yangxy/gpen:cf4e15a70049c0119884eb2906c8ae8807af8317bea98313fefd941e414d0c91",
        input={"image": open(image_path, "rb"), "task": task,
               "broken_image": broken_image}
    )  # type: ignore


def tmp(dir_input: str, dir_output: str):
    return 1



from typing import Any
from selenium.webdriver.common.by import By


def safe_find_element(parent: Any, by: str = By.ID,
                      value: str | None = None):
    try:
        return parent.find_element(by, value)
    except:
        return None



import logging
from typing import Any, Iterable
import pandas as pd

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)

formatter = logging.Formatter('%(asctime)s - %(levelname)-8s - %(message)s')
ch.setFormatter(formatter)
logger.addHandler(ch)

debug = logger.debug
info = logger.info
warning = logger.warning
error = logger.error
critical = logger.critical


def printTable(data: pd.DataFrame | dict[Any, Any] | Iterable[dict[Any, Any]]):
    df = pd.DataFrame(data)
    print(df)



from typing import Any

from material_zui.list import map_to, filter_to
from material_zui.string import not_empty


class ZuiBingAiResult():
    def __init__(self, response: dict[Any, Any]) -> None:
        self.response = response

    def get_valid_message(self, messages: list[str]) -> list[str]:
        return filter_to(messages, lambda text, _: not_empty(text))

    @property
    def full_texts(self) -> list[str]:
        texts: list[str] = map_to(self.response["item"]["messages"],
                                  lambda message, _: message.get('text') or message.get('hiddenText'))
        return self.get_valid_message(texts)

    @property
    def texts(self) -> list[str]:
        texts: list[str] = map_to(self.response["item"]["messages"],
                                  lambda message, _: message.get('text'))
        return self.get_valid_message(texts)

    @property
    def last_text(self) -> str:
        return self.texts.pop()



from urllib.request import urlopen


def download(url: str, output_path: str):
    mp4File = urlopen(url)
    with open(output_path, "wb") as output:
        while True:
            data = mp4File.read(4096)
            if data:
                output.write(data)
            else:
                break



# from material_zui.image.colorization import *
# from material_zui.image.combine import *
# from material_zui.image.convert import *
from material_zui.image.index import *
# from material_zui.image.upscale import *



import cv2

# Load the image
image = cv2.imread("static/img/chan-dung.jpg")

# Convert the image to grayscale
gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)

# Apply a threshold to the grayscale image
ret, thresh = cv2.threshold(gray, 127, 255, cv2.THRESH_BINARY)

# Find the contours of the objects in the thresholded image
contours, hierarchy = cv2.findContours(
    thresh, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)

# Find the largest contour in the image
largest_contour = max(contours, key=cv2.contourArea)

# Extract the bounding box of the largest contour
(x, y, w, h) = cv2.boundingRect(largest_contour)

# Crop the image to remove the background
cropped_image = image[y:y+h, x:x+w]

# Save the cropped image
cv2.imwrite("static/img/sketch/cropped_image.jpg", cropped_image)



from re import finditer, compile
from typing import Any, Callable

from material_zui.list import map_to, filter_to


# def list_match(value: str, regex: str, group: int = 0) -> list[str]:
#     pattern = compile(regex)
#     items = map_to(list(finditer(pattern, value)),
#                    lambda match, _: match.group(group))
#     return filter_to(items, lambda item, _: len(item) > 0)


def list_match(regex: str, group: int = 0) -> Callable[[str], list[str]]:
    pattern = compile(regex)

    def matcher(value: str) -> list[str]:
        items = map_to(list(finditer(pattern, value)),
                       lambda match, _: match.group(group))
        return filter_to(items, lambda item, _: len(item) > 0)
    return matcher

# def remove_all(text: str, regex_pattern_to_remove: str) -> str:
#     pattern = compile(regex_pattern_to_remove)
#     return pattern.sub('', text)

# def remove_all(regex_pattern_to_remove: str) -> Callable[[str], str]:
#     pattern = compile(regex_pattern_to_remove)
#     def remove_all_impl(text: str) -> str:
#         return pattern.sub('', text)
#     return remove_all_impl


def remove_all(regex_pattern_to_remove: str) -> Callable[[str], str]:
    pattern = compile(regex_pattern_to_remove)
    return lambda text: pattern.sub('', text)


def remove_special_characters(text: str) -> str:
    """
    The function removes all special characters from a given string.

    @param text: The input text string that may contain special characters
    @type text: str
    @return: The function `remove_special_characters` takes a string `text` as input and removes all
    special characters (i.e., non-alphanumeric characters) from it using a regular expression pattern.
    The function returns the modified string with special characters removed.
    """
    return remove_all('[^\\w\\s]')(text)


def remove_bmp_characters(text: str) -> str:
    """Removes all BMP characters from a string.
    Args:
      text: The string to remove BMP characters from.
    Returns:
      The string with all BMP characters removed.
    """
    return remove_all(r'[^\u0000-\uFFFF]')(text)


def trim_space(text: str) -> str:
    """
    The function takes a string as input, removes any leading or trailing spaces, and replaces any
    consecutive spaces within the string with a single space.

    :param text: A string that may contain extra spaces that need to be trimmed down to a single space
    :type text: str
    :return: The function `trim_space` takes a string `text` as input and returns a new string with all
    consecutive spaces replaced by a single space, and leading/trailing spaces removed.
    """
    pattern = compile('  ')
    return pattern.sub(' ', text.strip())


def not_empty(value: Any) -> bool:
    return value != None and len(value) > 0



import requests
import json
from material_zui.regex import value

url = "https://chatgpt.chatgptvietnam.org/api/chat-process"
headers = {
    'accept': 'application/json, text/plain, */*',
    'content-type': 'application/json',
    'origin': 'https://chatgpt.chatgptvietnam.org',
    'referer': 'https://chatgpt.chatgptvietnam.org/',
}


def gpt_call(prompt: str) -> str:
    payload = json.dumps({
        "prompt": prompt,
        "systemMessage": "You are ChatGPT, a large language model trained by OpenAI. Follow the user's instructions carefully. Respond using markdown.",
        "temperature": 0.8,
        "top_p": 1
    })
    response = requests.post(url, headers=headers, data=payload)
    return response.text


def gpt_last_result(prompt: str) -> dict[{'text': str}] | None:
    data = gpt_call(prompt)
    last_value = value(data, '.+"finish_reason":"stop".+')
    return json.loads(last_value) if last_value else None


def get_content(prompt: str) -> str:
    data = gpt_last_result(prompt)
    return data['text'] if data else ''



import datetime
from crontab import CronTab

# Create a new cron tab
my_cron = CronTab(user='chaunguyen')

# Create a new cron job
# job = my_cron.new(command='python3.10 /path/to/script.py')
job = my_cron.new(command='python3.10 src/material_zui_test/cron/test.py')

# Set the job to run every day at 5 PM
# job.setall('0 17 * * *')
# job.setall('1 * * * *')
job.minute.every(1)

print(job.hour, job.hours)
# my_cron.find_command("result")
print(my_cron.find_comment("result"))

for job in my_cron:
    sch = job.schedule(date_from=datetime.datetime.now())
    print(sch.get_next())

# Write the cron job to the cron tab
my_cron.write()



# import cv2
# import uuid
# from cv2 import Mat
# from material_zui.image.data import IMG_EXT, MAX_MP
# from material_zui.image.index import get_image_info, get_image_paths, get_images
# from material_zui.image.upscale import get_image_ratio_to_max_mp


# def save_upscale_toJpg(directory_path:  str, new_directory_path: str = '', max_mp: int = MAX_MP) -> None:
#     new_directory_path = new_directory_path if new_directory_path else directory_path
#     images = get_images(directory_path)
#     for index, image in enumerate(images):
#         ratio = get_image_ratio_to_max_mp(image, max_mp)
#         jpg_dir = f'{new_directory_path}/{index}.jpg'
#         upscale_image = cv2.resize(
#             image, None, fx=ratio, fy=ratio, interpolation=cv2.INTER_CUBIC)  # type: ignore
#         cv2.imwrite(jpg_dir, upscale_image)
#         print(index+1, jpg_dir, ratio)


# def handle_image(directory_path: str, new_directory_path: str = '', extension: str = ''):
#     image_info = get_image_info(directory_path)
#     dir_name = image_info['dir_name']
#     ext = extension or image_info['extension']
#     new_image_path = f'{dir_name}/{str(uuid.uuid4())}.{ext}'
#     new_directory_path = new_directory_path if new_directory_path else new_image_path
#     # print(new_directory_path)

#     def handle_each_image(*functions):
#         result: Mat = image_info['image']
#         for function in functions:
#             result = function(result)
#         cv2.imwrite(new_directory_path, result)
#     return handle_each_image


# def handle_images(directory_path: str, new_directory_path: str = '', extension: str = ''):
#     image_paths = get_image_paths(directory_path)

#     def handle_each_image(*functions):
#         for image_path in image_paths:
#             image_info = get_image_info(image_path)
#             name = image_info['name']
#             ext = extension or image_info['extension']
#             output_image_path = ''
#             if new_directory_path and directory_path != new_directory_path:
#                 output_image_path = f'{new_directory_path}/{name}.{ext}'
#             handle_image(image_path, output_image_path, extension)(*functions)
#     return handle_each_image


# # def save_upscale_toJpg(directory_path:  str, new_directory_path: str = '', max_mp: int = MAX_MP) -> None:
# #     new_directory_path = new_directory_path if new_directory_path else directory_path
# #     images = get_images(directory_path)
# #     image_paths = get_image_paths(directory_path)
# #     for index, image in enumerate(images):
# #         ratio = get_image_ratio_to_max_mp(image, max_mp)
# #         image_url = upscale(image_paths[index], ratio)
# #         png_dir = f'{new_directory_path}/{index}.png'
# #         jpg_dir = f'{new_directory_path}/{index}.jpg'
# #         download(image_url, png_dir)
# #         to_jpg(png_dir, jpg_dir)
# #         print(index+1, jpg_dir, ratio)



from uuid import uuid4
import requests
from bs4 import BeautifulSoup

from material_zui.image import download, download_base64, base64_ext
from material_zui.list import list_range
from material_zui.validate import is_base64, is_valid_url


def get_all_image_urls(url: str, limit: int = 0, start_index: int = 0) -> list[str]:
    '''
    Get all image url and base64 of image tags
    '''
    response = requests.get(url)
    soup = BeautifulSoup(response.content, "html.parser")
    images = soup.find_all("img")
    # Create a directory to store the downloaded images
    # if not os.path.exists("images"):
    #     os.mkdir("images")
    images = list_range(images, limit, start_index)
    return [image["src"] for image in images]


def download_image_urls(src_images: list[str], output_path: str) -> None:
    '''
    Download image from src url
    @src_images: must be a valid url (with `http(s)` prefix) or base64 string
    '''
    for src_image in src_images:
        if is_valid_url(src_image):
            print('Downloading', src_image)
            filename = src_image.split("/")[-1]
            file_output_path = f'{output_path}/{filename}'
            download(src_image, file_output_path)
        elif is_base64(src_image):
            print('Downloading', src_image)
            ext = base64_ext(src_image)
            if ext:
                filename = f'{str(uuid4())}.{ext}'
                file_output_path = f'{output_path}/{filename}'
                download_base64(src_image, file_output_path)
        else:
            print('Invalid type value', src_image)
    print("Download complete!")


def download_all_image(url: str, output_path: str, limit: int = 0, start_index: int = 0) -> None:
    '''
    Download all image from url
    @url: must be a valid url (with `http(s)` prefix)
    '''
    src_images = get_all_image_urls(url, limit, start_index)
    download_image_urls(src_images, output_path)

# def download_all_image(url: str, output_path: str, limit: int = 0, start_index: int = 0) -> None:
#     response = requests.get(url)
#     soup = BeautifulSoup(response.content, "html.parser")
#     images = soup.find_all("img")

#     # Create a directory to store the downloaded images
#     # if not os.path.exists("images"):
#     #     os.mkdir("images")
#     end_index = start_index+limit
#     images = get_all_image_urls(url,limit,start_index)
#     # src = "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAQAAAC1HAwCAAAAC0lEQVR42mNk+A8AAQUBAScY42YAAAAASUVORK5CYII="
#     # download_base64(src, f'{output_path}/abc.png')
#     # ext = base64_ext(src)
#     for image in images:
#         src = image["src"]
#         if is_valid_url(src):
#             print('Downloading', src)
#             filename = src.split("/")[-1]
#             file_output_path = f'{output_path}/{filename}'
#             download(src, file_output_path)
#         elif is_base64(src):
#             print('Downloading', src)
#             ext = base64_ext(src)
#             if ext:
#                 filename = f'{str(uuid4())}.{ext}'
#                 file_output_path = f'{output_path}/{filename}'
#                 download_base64(src, file_output_path)
#         else:
#             print('Invalid type value', src)
#     print("Download complete!")



import os
import sys
import shutil
from platform import platform


def is_mac() -> bool:
    '''
    Check current platform is `Mac` or not
    '''
    is_mac = False
    if not any(os_name in platform() for os_name in ["Windows", "Linux"]):
        is_mac = True
    return is_mac


def is_linux() -> bool: return sys.platform == "linux" or sys.platform == "linux2"


def is_window() -> bool: return sys.platform == "win64" or sys.platform == "win32"


def is_main() -> bool: return __name__ == "__main__"


def create_directory(directory_path: str) -> None:
    '''
    Check the exist of directory before create it
    '''
    if not os.path.exists(directory_path):
        os.mkdir(directory_path)


def move(source_file_path: str, destination_directory_path: str) -> None:
    shutil.move(source_file_path, destination_directory_path)



def encode(line):
    """
    INPUT:  A line of ASCII String.
    OUTPUT: A List of Run-Length Encoded of the input string.
    """
    if not line:
        return [""]
    else:
        last_char = line[0]
        length = len(line)
        i = 1
        while i < length and last_char == line[i]:
            i += 1
        return [str(i), last_char] + encode(line[i:])


def decode(parsed_string):
    """
    INPUT:  A List of List parsed from encoded string
    OUTPUT: A String of Decoded line
    """
    decoded = ""
    for item in parsed_string:
        try:
            decoded += item[0] * item[1]
        except IndexError:
            pass
    return decoded


def flatten(encoded_list):
    """
    INPUT:  A list of encoded ASCII String
    OUPUT:  A String of encoded ASCII String
    """
    return "".join(encoded_list)


def parse(string):
    """
    This function parses the encoded string for the decoding phase.
    This is especially userful when a character shows up
    more than nine times in a row.

    INPUT:  An Encoded String
    OUTPUT: A List of List // Parsed Encoded String
    """
    if not string or len(string) == 1:
        return [""]
    if string[1].isdigit():
        return [[int(string[:2]), string[2]]] + parse(string[3:])
    else:
        return [[int(string[0]), string[1]]] + parse(string[2:])


def save_text_encode(file_path: str, output_path: str):
    '''
    @file_path: `abc/def.txt`
    @output_path: `ghk/ilm.txt`
    - Only apply for `text file (*.txt)`
    - Use `lossless compression algorithm`
    - Text file after encode will reduce volume size
    '''
    with open(file_path, 'r') as f:
        with open(output_path, 'w+') as o:
            for line in f:
                encoded_string = flatten((encode(line)))
                o.write(encoded_string)


def save_text_decode(file_path: str, output_path: str):
    '''
    Only apply for text file (*.txt)
    '''
    with open(file_path, 'r') as f:
        with open(output_path, 'w+') as o:
            for line in f:
                parsed_string = parse(line)
                o.write(decode(parsed_string))



from typing import Any
import xlsxwriter

from .type import ZuiExcelColumns

# Create a workbook and add a worksheet.
# workbook = xlsxwriter.Workbook('static/file/demo.xlsx')
# worksheet = workbook.add_worksheet()

# # Add a bold format to use to highlight cells.
# bold = workbook.add_format({'bold': 1})

# # Add a number format for cells with money.
# money_format = workbook.add_format({'num_format': '$#,##0'})

# # Add an Excel date format.
# date_format = workbook.add_format({'num_format': 'mmmm d yyyy'})

# # Adjust the column width.
# worksheet.set_column(1, 1, 15)

# # Write some data headers.
# # worksheet.write('A1', 'Item', bold)
# worksheet.write(0, 0, 'Item2', bold)
# worksheet.write('B1', 'Date', bold)
# worksheet.write('C1', 'Cost', bold)

# # Some data we want to write to the worksheet.
# expenses = (
#     ['Rent', '2013-01-13', 1000],
#     ['Gas',  '2013-01-14',  100],
#     ['Food', '2013-01-16',  300],
#     ['Gym',  '2013-01-20',   50],
# )

# # Start from the first cell below the headers.
# row = 1
# col = 0

# for item, date_str, cost in (expenses):
#     # Convert the date string into a datetime object.
#     date = datetime.strptime(str(date_str), "%Y-%m-%d")

#     worksheet.write_string(row, col,     item)
#     worksheet.write_datetime(row, col + 1, date, date_format)
#     worksheet.write_number(row, col + 2, cost, money_format)
#     row += 1

# # Write a total using a formula.
# worksheet.write(row, 0, 'Total', bold)
# worksheet.write(row, 2, '=SUM(C2:C5)', money_format)

# workbook.close()


def write_excel(file_path: str, columns: ZuiExcelColumns, data: list[dict[str, Any]]) -> None:
    workbook = xlsxwriter.Workbook(file_path)
    worksheet = workbook.add_worksheet()

    for index, column in enumerate(columns):
        worksheet.write(0, index, column['name'])

    for rowIndex, item in enumerate(data):
        for colIndex, column in enumerate(columns):
            value = item.get(column['field'], '')
            worksheet.write(rowIndex+1, colIndex, value)
    workbook.close()



import os
from bardapi import Bard


def set_key(key: str):
    '''
    - set your `__Secure-1PSID` value to key
    - detail info: https://github.com/dsdanielpark/Bard-API
    - amazing prompt: https://github.com/dsdanielpark/amazing-bard-prompts
    '''
    os.environ['_BARD_API_KEY'] = key


def get_response(
    prompt: str) -> dict[{'content': str, 'choices': list[dict[{'id': str, 'content': list[str]}]]}]: return Bard().get_answer(prompt)  # type: ignore


def get_content(prompt: str) -> str: return get_response(prompt)['content']



import cv2
from cv2 import Mat
from rembg import remove

# with open(input_path, 'rb') as i:
#     with open(output_path, 'wb') as o:
#         input = i.read()
#         output = remove(input)
#         o.write(output)  # type: ignore


def transparent_background(
    image: Mat) -> Mat: return remove(image)  # type: ignore


def save_transparent_background(image_path: str, output_path: str) -> None:
    input = cv2.imread(image_path)
    output = transparent_background(input)
    cv2.imwrite(output_path, output)



from setuptools import setup
import setuptools

# with open("readme.md", "r") as fh:
#     long_description = fh.read()
with open("src/material_zui/readme.md", "r") as fh:
    long_description = fh.read()

setup(
    name='material_zui',
    version='0.0.6',
    # url='https://github.com/username/my_module',
    author='chauhmnguyen',
    author_email='chauhoangminhnguyen@gmail.com',
    description='Material Zui',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=setuptools.find_packages(),
    install_requires=['cv2', 'PIL'],
    classifiers=[
        "Operating System :: OS Independent",
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
)



import validators
from base64 import b64decode


def is_valid_url(url) -> bool: return validators.url(url)  # type: ignore


def is_base64(value: str) -> bool:
    try:
        data = value.split(',')[1]
        decoded_img = b64decode(data)
        return bool(decoded_img)
    except:
        return False



import telebot


class zui_bot(telebot.TeleBot):
    def __init__(self, token: str):
        super().__init__(token)

    def send_file(self, chat_id: int | str, file_path: str):
        with open(file_path, "rb") as f:
            super().send_document(chat_id, f)


def create_bot(token: str):
    '''
    Detail API: https://pypi.org/project/pyTelegramBotAPI
    '''
    # return telebot.TeleBot(token)
    return zui_bot(token)



import threading
import time


def set_timeout(code, milliseconds):
    timer = threading.Timer(milliseconds / 1000, code)
    timer.start()
    return timer


def clear_timeout(timer):
    timer.cancel()


def my_code():
    print("Hello World!")


timer = set_timeout(my_code, 5)


def set_interval(func, sec):
    def func_wrapper():
        set_interval(func, sec)
        func()
    # t = threading.Timer(sec, func_wrapper)
    t = threading.Timer(sec, func)
    t.start()
    return t


def clear_interval(t):
    print('cancel')
    t.cancel()


def print_hello():
    print("Hello!")


interval_id = set_interval(print_hello, 1)
time.sleep(5)
clear_interval(interval_id)
print('abc')



from datetime import datetime, timedelta


def now() -> datetime: return datetime.now()


def add_days(date: datetime, days: int) -> datetime:
    diff = timedelta(days=days)
    return date + diff


def to_date(date: str) -> datetime:
    return datetime.strptime(date, "%Y-%m-%d")


def to_date_str(date_time: datetime) -> str:
    return datetime.strftime(date_time, "%Y-%m-%d")


def to_date_time(date_time: str) -> datetime:
    return datetime.strptime(date_time, "%Y-%m-%d %H:%M:%S")



from typing import Callable, TypeVar
from functools import reduce

from material_zui.list import map_to

T = TypeVar("T")
R = TypeVar("R")


def pipe(
    func: Callable[[T], R], *funcs: Callable[[T], R]
) -> Callable[[T], R]:
    """Composes multiple functions into a single function.
    Args:
        func: The first function to compose.
        funcs: The remaining functions to compose.

    Returns:
        A function that applies all of the composed functions to its input.
    """
    def composed(x: T) -> R:
        return reduce(lambda x, f: f(x), (func, *funcs), x)  # type: ignore
    return composed  # type: ignore


def pipe_list(
    func: Callable[[T], R], *funcs: Callable[[T], R]
) -> Callable[[list[T]], list[R]]:
    def composed(items: list[T]) -> list[R]:
        return map_to(items, lambda item, _: pipe(func, *funcs)(item))
    return composed  # type: ignore



from PIL import Image
import os
import cv2
from cv2 import Mat
from material_zui.image.data import K_SIZE, MAX_MP
from material_zui.log.index import printTable

# def convert_to_sketch(image: Mat) -> Mat:
#     gray_image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
#     inverted_gray_image = 255 - gray_image  # type: ignore
#     blurred_image = cv2.GaussianBlur(inverted_gray_image, (21, 21), 0)
#     inverted_blurred_image = 255 - blurred_image
#     pencil_sketch = cv2.divide(
#         gray_image, inverted_blurred_image, scale=256.0)
#     return pencil_sketch


# def to_sketch(img: Mat, k_size: int = K_SIZE) -> Mat:
#     grey_img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
#     invert_img = cv2.bitwise_not(grey_img)
#     blur_img = cv2.GaussianBlur(invert_img, (k_size, k_size), 0)
#     invblur_img = cv2.bitwise_not(blur_img)
#     return cv2.divide(grey_img, invblur_img, scale=256.0)

def to_sketch(k_size: int = K_SIZE):
    def handle_image(image: Mat):
        grey_img = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        invert_img = cv2.bitwise_not(grey_img)
        blur_img = cv2.GaussianBlur(invert_img, (k_size, k_size), 0)
        invblur_img = cv2.bitwise_not(blur_img)
        return cv2.divide(grey_img, invblur_img, scale=256.0)
    return handle_image


# def save_sketch_then_upscale(directory_path:  str, new_directory_path: str = '', k_size: int = 21, max_mp: int = MAX_MP) -> None:
#     images = upscales_dir_to_max_size(
#         directory_path, new_directory_path, max_mp)
#     for image in images:
#         sketch_image = to_sketch(k_size)(image['matImage'])
#         cv2.imwrite(os.path.join(new_directory_path,
#                     image['name']), sketch_image)
#     printTable({
#         'Name': [info['name'] for info in images],
#         'Width': [info['width'] for info in images],
#         'Height': [info['height'] for info in images],
#         'Ratio': [info['ratio'] for info in images],
#     })


# def to_jpg(image: Mat) -> None:
#     '''
#     @image_path: `abc/def/input.png`
#     @output_path: `abc/def/output.jpg` or `abc/def/output.jpeg`
#     '''
#     # image = Image.open(image_path)
#     # image.save(output_path, format="JPEG")
#     # image.close()
#     # cv2.imread(image_path)
#     # image.to


def save_to_jpg(image_path: str, output_path: str) -> None:
    '''
    @image_path: `abc/def/input.png`
    @output_path: `abc/def/output.jpg` or `abc/def/output.jpeg`
    '''
    image = Image.open(image_path)
    image.save(output_path, format="JPEG")
    image.close()


def to_grayscale(image: Mat): return cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)


def save_to_grayscale(image_path: str, output_path: str):
    image = cv2.imread(image_path)
    grayscale_image = to_grayscale(image)
    cv2.imwrite(output_path, grayscale_image)



import random
from time import sleep


def int_number(min: int, max: int) -> int: return random.randint(min, max)


def random_sleep(min: int = 1, max: int = 5, sec: float = 0) -> None:
    time = sec if sec > 0 else int_number(min, max)
    sleep(time)



import os
import re
import uuid
from base64 import b64decode
from pathlib import Path
import cv2
import requests
from cv2 import Mat

from material_zui.image.data import IMG_EXT
from material_zui.log import printTable

# def save_upscale_toJpg(directory_path:  str, new_directory_path: str = '', max_mp: int = MAX_MP) -> None:
#     new_directory_path = new_directory_path if new_directory_path else directory_path
#     images = get_images(directory_path)
#     for index, image in enumerate(images):
#         ratio = get_image_ratio_to_max_mp(image, max_mp)
#         jpg_dir = f'{new_directory_path}/{index}.jpg'
#         upscale_image = cv2.resize(
#             image, None, fx=ratio, fy=ratio, interpolation=cv2.INTER_CUBIC)  # type: ignore
#         cv2.imwrite(jpg_dir, upscale_image)
#         print(index+1, jpg_dir, ratio)


def handle_image(directory_path: str, new_directory_path: str = '', extension: str = ''):
    image_info = get_image_info(directory_path)
    dir_name = image_info['dir_name']
    ext = extension or image_info['extension']
    new_image_path = f'{dir_name}/{str(uuid.uuid4())}.{ext}'
    new_directory_path = new_directory_path if new_directory_path else new_image_path

    def handle_each_image(*functions):
        result: Mat = image_info['image']
        for function in functions:
            result = function(result)
        cv2.imwrite(new_directory_path, result)
    return handle_each_image


def handle_images(directory_path: str, new_directory_path: str = '', extension: str = ''):
    image_paths = get_image_paths(directory_path)

    def handle_each_image(*functions):
        items = []
        for image_path in image_paths:
            image_info = get_image_info(image_path)
            name = image_info['name']
            ext = extension or image_info['extension']
            output_image_path = ''
            if new_directory_path and directory_path != new_directory_path:
                output_image_path = f'{new_directory_path}/{name}.{ext}'
            handle_image(image_path, output_image_path, extension)(*functions)
            items.append((image_path, output_image_path))
        printTable({
            'Input path': [item[0] for item in items],
            'Output path': [item[1] for item in items],
        })
    return handle_each_image

# def get_image_sizes(directory_path: str) -> list[dict[{'name': str, 'width': int, 'height': int}]]:
#     images = get_image(directory_path)
#     images_name = get_image_name(directory_path)
#     return [{'name': images_name[index], 'width':image.width, 'height':image.height}
#             for index, image in enumerate(images)]


# def get_image(directory_path: str) -> list[Image.Image]:
#     images_name = get_image_name(directory_path)
#     # print(images_name)
#     return list(map(lambda file_name: Image.open(os.path.join(
#         directory_path, file_name)), images_name))
# def get_image(directory_path: str) -> list[ZuiImage]:
#     images_name = get_image_name(directory_path)
#     images = [Image.open(os.path.join(directory_path, file_name))
#               for file_name in images_name]
#     return [{'image': image, 'name': images_name[index], 'ext': image.format or '', 'width':image.width, 'height':image.height} for index, image in enumerate(images)]


def get_image_names(directory_path: str) -> list[str]: return list(filter(lambda file_name: file_name.endswith(
    IMG_EXT), os.listdir(directory_path)))


def get_image_paths(directory_path: str) -> list[str]:
    images_name = get_image_names(directory_path)
    return [os.path.join(directory_path, image_name) for image_name in images_name]


def get_images(directory_path: str) -> list[Mat]: return [cv2.imread(
    image_path) for image_path in get_image_paths(directory_path)]


# def get_image_info(image_path: str) -> dict[{'image': Mat, 'name': str, 'dir_name': str, 'width': int, 'height': int, 'extension': str}]:
#     '''
#     @file_name: `abc.jpg`
#     @name: `abc`
#     @extension: `jpg`
#     '''
#     image = cv2.imread(image_path)
#     height, width = image.shape[:2]
#     file_name = os.path.basename(image_path)
#     name = os.path.basename(image_path)
#     extension = os.path.splitext(image_path)[1][1:]
#     dir_name = os.path.dirname(image_path)
#     # print(width, height, extension)
#     return {'image': image, 'name': name, 'dir_name': dir_name, 'width': width, 'height': height, 'extension': extension}

def get_image_info(image_path: str) -> dict[{'image': Mat, 'file_name': str, 'name': str, 'dir_name': str, 'width': int, 'height': int, 'extension': str}]:
    '''
    @file_name: `abc.jpg`
    @name: `abc`
    @extension: `jpg`
    '''
    path = Path(image_path)
    image = cv2.imread(image_path)
    height, width = image.shape[:2]
    file_name = path.name
    name = path.stem
    extension = os.path.splitext(image_path)[1][1:]
    dir_name = os.path.dirname(image_path)
    return {'image': image, 'file_name': file_name, 'name': name, 'dir_name': dir_name, 'width': width, 'height': height, 'extension': extension}


def get_image_size(image: Mat) -> tuple[int, int]:
    height, width = image.shape[:2]
    return (width, height)

# def change_image_path(file_path: str, file_name: str) -> str:
#     path_prefix = os.path.splitext(file_path)[0]
#     return f'{path_prefix}{file_name}'


def download(url: str, output_path: str) -> None:
    '''
    Download image by url to path
    @url: must be a valid url (with `http(s)` prefix)
    @output_path: include file name
        - ex: `abc/def.png`
    '''
    response = requests.get(url)
    with open(output_path, "wb") as f:
        f.write(response.content)


# def download_base64(base64_image: str, output_path: str) -> None:
#     '''
#     Download image with base64 data input
#     @url: with format `data:image/{extension};base64,...`
#     @output_path: include file name
#         - ex: `abc/def.png`
#     '''
#     # base64_image = "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAEYAAAAUCAYAAACNiR0NAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAAAJcEhZcwAADsQAAA7EAZUrDhsAAAANSURBVBhXYzh8+PB/AAffA0nNPuCLAAAAAElFTkSuQmCC"
#     image_data = base64.b64decode(base64_image)
#     image = Image.open(io.BytesIO(image_data))
#     image.save(output_path)


def download_base64(base64_image: str, output_path: str) -> None:
    '''
    Download image with base64 data input
    @base64_image: with format `data:image/{extension};base64,...`
    @output_path: include file name
        - ex: `abc/def.png`
    '''
    data = base64_image.split(',')[1]
    decoded_img = b64decode(data)
    with open(output_path, 'wb') as f:
        f.write(decoded_img)


def base64_ext(base64_image: str) -> str | None:
    '''
    Get extension of base64 string data
    @base64_image: with format `data:image/{extension};base64,...`
    @ex:
        - input: 
        - ex: `abc/def.png`
    '''
    pattern = r"data:image/(\w+);base64"
    match = re.search(pattern, base64_image)
    if match:
        return match.group(1)
    else:
        return None



from PIL import Image
from cv2 import Mat

ZuiImage = dict[{'name': str, 'ext': str, 'width': int, 'height': int,
                 'image': Image.Image, 'matImage': Mat,  'ratio': int}]

ZuiImageRes = dict[str, str | int | Image.Image | Mat | None]



import os
from langchain import PromptTemplate, HuggingFaceHub, LLMChain, HuggingFacePipeline

from material_zui.bing_ai.bing_ai import ZuiBingAi


class ZuiGen:
    def __init__(self, token: str, cookie_dir_path: str) -> None:
        self.token = os.environ['HUGGINGFACEHUB_API_TOKEN'] = token
        self.bing_ai = ZuiBingAi(cookie_dir_path)

    def gen_title(self, topic: str) -> str:
        prompt = topic
        # data = self.bing_ai.query(prompt)
        data = self.bing_ai.query(
            'Write seven subheadings for the blog article with the title yoga; the titles should be catchy and 60 characters max.')
        # print(data.response)
        # print(data.full_texts)
        print(data.texts)
        print(data.last_text)
        return ''

    def gen_titles(self, topics: list[str]) -> list[str]:
        return []

    def test(self):
        template = """
        Question: {question}
        Answer: """

        prompt = PromptTemplate(
            template=template,
            input_variables=['question']
        )
        question = "Which NFL team won the Super Bowl in the 2010 season?"

        # initialize Hub LLM
        hub_llm = HuggingFaceHub(
            # repo_id='google/flan-t5-xl',
            repo_id='OpenAssistant/oasst-sft-6-llama-30b-xor',
            model_kwargs={'temperature': 1e-10}
        )

        # create prompt template > LLM chain
        llm_chain = LLMChain(
            prompt=prompt,
            llm=hub_llm
        )

        # ask the user question about NFL 2010
        print(llm_chain.run(question))

    def ask(self, ask: str) -> str:
        template = """Question: {query}
        Answer: Let's think step by step."""
        prompt = PromptTemplate(template=template, input_variables=["query"])
        llm = HuggingFaceHub(repo_id="google/flan-t5-xxl",
                             model_kwargs={"temperature": 0.1})  # type: ignore
        llm_chain = LLMChain(prompt=prompt, llm=llm)
        # query = "When was America discovered?"
        # query = "What the nodejs latest version?"
        return llm_chain.run(ask)



from material_zui.image.colorization import colorization, save_colorization
from material_zui.image.common import download, handle_image, handle_images, download_base64, base64_ext
from material_zui.image.convert import (save_to_grayscale, save_to_jpg,
                                        to_grayscale, to_sketch)
from material_zui.image.transparent_background import (
    save_transparent_background, transparent_background)
from material_zui.image.upscale import upscale_to_max_size
